@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2019 Jeffrey Stedfast
+ *  Copyright (C) 2000-2020 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -263,8 +263,11 @@ locale_parse_lang (const char *locale)
 void
 g_mime_charset_map_init (void)
 {
-	char *charset, *iconv_name, *locale;
+	char *charset, *iconv_name;
 	int i;
+#ifndef WIN32
+	char *locale;
+#endif
 	
 	initialized = MAX (initialized, 0);
 	
